@@ -62,7 +62,7 @@ def _rescan_device(device):
                    kernel partition records.
     """
     try:
-        utils.execute('partx', '-u', device, attempts=3,
+        utils.execute('partx', '-a', device, attempts=3,
                       delay_on_retry=True)
         utils.execute('udevadm', 'settle')
     except processutils.ProcessExecutionError:
@@ -273,16 +273,16 @@ def _run_efibootmgr(valid_efi_bootloaders, device, efi_partition,
     """
 
     # Before updating let's get information about the bootorder
-    LOG.debug("Getting information about boot order")
-    utils.execute('efibootmgr')
-    # NOTE(iurygregory): regex used to identify the Warning in the stderr after
-    # we add the new entry. Example:
-    # "efibootmgr: ** Warning ** : Boot0004 has same label ironic"
-    duplicated_label = re.compile(r'^.*:\s\*\*.*\*\*\s:\s.*'
-                                  r'Boot([0-9a-f-A-F]+)\s.*$')
+    LOG.debug("Getting information about boot order.")
+    original_efi_output = utils.execute('efibootmgr', '-v')
+    # NOTE(TheJulia): regex used to identify entries in the efibootmgr
+    # output on stdout.
+    entry_label = re.compile(r'Boot([0-9a-f-A-F]+)\*?\s(.*).*$')
     label_id = 1
     for v_bl in valid_efi_bootloaders:
         if 'csv' in v_bl.lower():
+            LOG.debug('A CSV file has been identified as a bootloader hint. '
+                      'File: %s', v_bl)
             # These files are always UTF-16 encoded, sometimes have a header.
             # Positive bonus is python silently drops the FEFF header.
             with open(mount_point + '/' + v_bl, 'r', encoding='utf-16') as csv:
@@ -296,23 +296,60 @@ def _run_efibootmgr(valid_efi_bootloaders, device, efi_partition,
             v_efi_bl_path = '\\' + v_bl.replace('/', '\\')
             label = 'ironic' + str(label_id)
 
+        # Iterate through standard out, and look for duplicates
+        for line in original_efi_output[0].split('\n'):
+            match = entry_label.match(line)
+            # Look for the base label in the string if a line match
+            # occurs, so we can identify if we need to eliminate the
+            # entry.
+            if match and label in match.group(2):
+                boot_num = match.group(1)
+                LOG.debug("Found bootnum %s matching label", boot_num)
+                utils.execute('efibootmgr', '-b', boot_num, '-B')
+
         LOG.debug("Adding loader %(path)s on partition %(part)s of device "
                   " %(dev)s", {'path': v_efi_bl_path, 'part': efi_partition,
                                'dev': device})
         # Update the nvram using efibootmgr
         # https://linux.die.net/man/8/efibootmgr
-        cmd = utils.execute('efibootmgr', '-c', '-d', device,
-                            '-p', efi_partition, '-w', '-L', label,
-                            '-l', v_efi_bl_path)
-        for line in cmd[1].split('\n'):
-            match = duplicated_label.match(line)
-            if match:
-                boot_num = match.group(1)
-                LOG.debug("Found bootnum %s matching label", boot_num)
-                utils.execute('efibootmgr', '-b', boot_num, '-B')
+        utils.execute('efibootmgr', '-v', '-c', '-d', device,
+                      '-p', str(efi_partition), '-w', '-L', label,
+                      '-l', v_efi_bl_path)
+        # Increment the ID in case the loop runs again.
         label_id += 1
 
 
+def get_partition_path_by_number(device, part_num):
+    """Get partition path (/dev/something) by a partition number on device.
+
+    Only works for GPT partition table.
+    """
+    uuid = None
+    partinfo, _ = utils.execute('sgdisk', '-i', str(part_num), device,
+                                use_standard_locale=True)
+    for line in partinfo.splitlines():
+        if not line.strip():
+            continue
+
+        try:
+            field, value = line.rsplit(':', 1)
+        except ValueError:
+            LOG.warning('Invalid sgdisk line: %s', line)
+            continue
+
+        if 'partition unique guid' in field.lower():
+            uuid = value.strip().lower()
+            LOG.debug('GPT partition number %s on device %s has UUID %s',
+                      part_num, device, uuid)
+            break
+
+    if uuid is not None:
+        return _get_partition(device, uuid)
+    else:
+        LOG.warning('No UUID information provided in sgdisk output for '
+                    'partition %s on device %s', part_num, device)
+
+
 def _manage_uefi(device, efi_system_part_uuid=None):
     """Manage the device looking for valid efi bootloaders to update the nvram.
 
@@ -328,6 +365,7 @@ def _manage_uefi(device, efi_system_part_uuid=None):
     """
     efi_partition_mount_point = None
     efi_mounted = False
+    efi_partition = None
 
     try:
         # Force UEFI to rescan the device.
@@ -335,17 +373,27 @@ def _manage_uefi(device, efi_system_part_uuid=None):
 
         local_path = tempfile.mkdtemp()
         # Trust the contents on the disk in the event of a whole disk image.
-        efi_partition = utils.get_efi_part_on_device(device)
+        efi_part_num = utils.get_efi_part_on_device(device)
+        if efi_part_num is not None:
+            efi_partition = get_partition_path_by_number(device, efi_part_num)
+
         if not efi_partition and efi_system_part_uuid:
             # _get_partition returns <device>+<partition> and we only need the
             # partition number
-            partition = _get_partition(device, uuid=efi_system_part_uuid)
+            efi_partition = _get_partition(device, uuid=efi_system_part_uuid)
             try:
-                efi_partition = int(partition.replace(device, ""))
+                efi_part_num = int(efi_partition.replace(device, ""))
             except ValueError:
                 # NVMe Devices get a partitioning scheme that is different from
                 # traditional block devices like SCSI/SATA
-                efi_partition = int(partition.replace(device + 'p', ""))
+                try:
+                    efi_part_num = int(efi_partition.replace(device + 'p', ""))
+                except ValueError as exc:
+                    # At least provide a reasonable error message if the device
+                    # does not follow this procedure.
+                    raise errors.DeviceNotFound(
+                        "Cannot detect the partition number of the device "
+                        "%s: %s" % (device, exc))
 
         if not efi_partition:
             # NOTE(dtantsur): we cannot have a valid EFI deployment without an
@@ -362,25 +410,18 @@ def _manage_uefi(device, efi_system_part_uuid=None):
         if not os.path.exists(efi_partition_mount_point):
             os.makedirs(efi_partition_mount_point)
 
-        # The mount needs the device with the partition, in case the
-        # device ends with a digit we add a `p` and the partition number we
-        # found, otherwise we just join the device and the partition number
-        if device[-1].isdigit():
-            efi_device_part = '{}p{}'.format(device, efi_partition)
-            utils.execute('mount', efi_device_part, efi_partition_mount_point)
-        else:
-            efi_device_part = '{}{}'.format(device, efi_partition)
-            utils.execute('mount', efi_device_part, efi_partition_mount_point)
+        utils.execute('mount', efi_partition, efi_partition_mount_point)
         efi_mounted = True
 
         valid_efi_bootloaders = _get_efi_bootloaders(efi_partition_mount_point)
         if valid_efi_bootloaders:
-            _run_efibootmgr(valid_efi_bootloaders, device, efi_partition,
+            _run_efibootmgr(valid_efi_bootloaders, device, efi_part_num,
                             efi_partition_mount_point)
             return True
         else:
             # NOTE(dtantsur): if we have an empty EFI partition, try to use
             # grub-install to populate it.
+            LOG.warning('Empty EFI partition detected.')
             return False
 
     except processutils.ProcessExecutionError as e:
@@ -551,6 +592,29 @@ def _umount_all_partitions(path, path_variable, umount_warn_msg):
     return umount_binds_success
 
 
+def _mount_partition(partition, path):
+    if not os.path.ismount(path):
+        LOG.debug('Attempting to mount %(device)s to %(path)s to '
+                  'partition.',
+                  {'device': partition,
+                   'path': path})
+        try:
+            utils.execute('mount', partition, path)
+        except processutils.ProcessExecutionError as e:
+            # NOTE(TheJulia): It seems in some cases,
+            # the python os.path.ismount can return False
+            # even *if* it is actually mounted. This appears
+            # to be becasue it tries to rely on inode on device
+            # logic, yet the rules are sometimes different inside
+            # ramdisks. So lets check the error first.
+            if 'already mounted' not in e:
+                # Raise the error, since this is not a known
+                # failure case
+                raise
+            else:
+                LOG.debug('Partition already mounted, proceeding.')
+
+
 def _install_grub2(device, root_uuid, efi_system_part_uuid=None,
                    prep_boot_part_uuid=None, target_boot_mode='bios'):
     """Install GRUB2 bootloader on a given device."""
@@ -638,11 +702,9 @@ def _install_grub2(device, root_uuid, efi_system_part_uuid=None,
                 # remounted.
                 LOG.debug('No EFI assets were preserved for setup or the '
                           'ramdisk was unable to complete the setup. '
-                          'falling back to bootloader installation from'
+                          'falling back to bootloader installation from '
                           'deployed image.')
-                if not os.path.ismount(path):
-                    LOG.debug('Re-mounting the root partition.')
-                    utils.execute('mount', root_partition, path)
+                _mount_partition(root_partition, path)
 
         binary_name = "grub"
         if os.path.exists(os.path.join(path, 'usr/sbin/grub2-install')):
@@ -662,36 +724,41 @@ def _install_grub2(device, root_uuid, efi_system_part_uuid=None,
             LOG.warning("GRUB2 will be installed for UEFI on efi partition "
                         "%s using the install command which does not place "
                         "Secure Boot signed binaries.", efi_partition)
-            utils.execute('mount', efi_partition, efi_partition_mount_point)
+
+            _mount_partition(efi_partition, efi_partition_mount_point)
             efi_mounted = True
-            # FIXME(rg): does not work in cross boot mode case (target
-            # boot mode differs from ramdisk one)
-            # Probe for the correct target (depends on the arch, example
-            # --target=x86_64-efi)
-            utils.execute('chroot %(path)s /bin/sh -c '
-                          '"%(bin)s-install"' %
-                          {'path': path, 'bin': binary_name},
-                          shell=True,
-                          env_variables={
-                              'PATH': path_variable
-                          })
-            # Also run grub-install with --removable, this installs grub to
-            # the EFI fallback path. Useful if the NVRAM wasn't written
-            # correctly, was reset or if testing with virt as libvirt
-            # resets the NVRAM on instance start.
-            # This operation is essentially a copy operation. Use of the
-            # --removable flag, per the grub-install source code changes
-            # the default file to be copied, destination file name, and
-            # prevents NVRAM from being updated.
-            # We only run grub2_install for uefi if we can't verify the
-            # uefi bits
-            utils.execute('chroot %(path)s /bin/sh -c '
-                          '"%(bin)s-install --removable"' %
-                          {'path': path, 'bin': binary_name},
-                          shell=True,
-                          env_variables={
-                              'PATH': path_variable
-                          })
+            try:
+                utils.execute('chroot %(path)s /bin/sh -c '
+                              '"%(bin)s-install"' %
+                              {'path': path, 'bin': binary_name},
+                              shell=True,
+                              env_variables={
+                                  'PATH': path_variable
+                              })
+            except processutils.ProcessExecutionError as e:
+                LOG.warning('Ignoring GRUB2 boot loader installation failure: '
+                            '%s.', e)
+            try:
+                # Also run grub-install with --removable, this installs grub to
+                # the EFI fallback path. Useful if the NVRAM wasn't written
+                # correctly, was reset or if testing with virt as libvirt
+                # resets the NVRAM on instance start.
+                # This operation is essentially a copy operation. Use of the
+                # --removable flag, per the grub-install source code changes
+                # the default file to be copied, destination file name, and
+                # prevents NVRAM from being updated.
+                # We only run grub2_install for uefi if we can't verify the
+                # uefi bits
+                utils.execute('chroot %(path)s /bin/sh -c '
+                              '"%(bin)s-install --removable"' %
+                              {'path': path, 'bin': binary_name},
+                              shell=True,
+                              env_variables={
+                                  'PATH': path_variable
+                              })
+            except processutils.ProcessExecutionError as e:
+                LOG.warning('Ignoring GRUB2 boot loader installation failure: '
+                            '%s.', e)
             utils.execute('umount', efi_partition_mount_point, attempts=3,
                           delay_on_retry=True)
             efi_mounted = False
@@ -730,7 +797,7 @@ def _install_grub2(device, root_uuid, efi_system_part_uuid=None,
         # recommended path.
         _configure_grub(device, path)
 
-        if efi_mounted:
+        if efi_system_part_uuid and efi_mounted:
             _append_uefi_to_fstab(path, efi_system_part_uuid)
 
         LOG.info("GRUB2 successfully installed on %s", device)
@@ -883,17 +950,58 @@ def _try_preserve_efi_assets(device, path,
 def _append_uefi_to_fstab(fs_path, efi_system_part_uuid):
     """Append the efi partition id to the filesystem table.
 
-    :param fs_path:
-    :param efi_system_part_uuid:
+    :param fs_path: The path to the filesystem.
+    :param efi_system_part_uuid: uuid to use to try and find the
+                                 partition. Warning: this may be
+                                 a partition uuid or a actual uuid.
     """
     fstab_file = os.path.join(fs_path, 'etc/fstab')
     if not os.path.exists(fstab_file):
         return
     try:
-        fstab_string = ("UUID=%s\t/boot/efi\tvfat\tumask=0077\t"
-                        "0\t1\n") % efi_system_part_uuid
+        # Collect all of the block devices so we appropriately match UUID
+        # or PARTUUID into an fstab entry.
+        block_devs = hardware.list_all_block_devices(block_type='part')
+
+        # Default to uuid, but if we find a partuuid instead, that is okay,
+        # we just need to know later on.
+        fstab_label = None
+        for bdev in block_devs:
+            # Check UUID first
+            if bdev.uuid and efi_system_part_uuid in bdev.uuid:
+                LOG.debug('Identified block device %(dev)s UUID %(uuid)s '
+                          'for UEFI boot. Proceeding with fstab update using '
+                          'a UUID.',
+                          {'dev': bdev.name,
+                           'uuid': efi_system_part_uuid})
+                # What we have works, and is correct, we can break the loop
+                fstab_label = 'UUID'
+                break
+            # Fallback to PARTUUID, since we don't know if the provided
+            # UUID matches a PARTUUID, or UUID field, and the fstab entry
+            # needs to match it.
+            if bdev.partuuid and efi_system_part_uuid in bdev.partuuid:
+                LOG.debug('Identified block device %(dev)s partition UUID '
+                          '%(uuid)s for UEFI boot. Proceeding with fstab '
+                          'update using a PARTUUID.',
+                          {'dev': bdev.name,
+                           'uuid': efi_system_part_uuid})
+                fstab_label = 'PARTUUID'
+                break
+
+        if not fstab_label:
+            # Fallback to prior behavior, which should generally be correct.
+            LOG.warning('Falling back to fstab entry addition label of UUID. '
+                        'We could not identify which UUID or PARTUUID '
+                        'identifier label should be used, thus UUID will be '
+                        'used.')
+            fstab_label = 'UUID'
+
+        fstab_string = ("%s=%s\t/boot/efi\tvfat\tumask=0077\t"
+                        "0\t1\n") % (fstab_label, efi_system_part_uuid)
         with open(fstab_file, "r+") as fstab:
-            if efi_system_part_uuid not in fstab.read():
+            already_present_string = fstab_label + '=' + efi_system_part_uuid
+            if already_present_string not in fstab.read():
                 fstab.writelines(fstab_string)
     except (OSError, EnvironmentError, IOError) as exc:
         LOG.debug('Failed to add entry to /etc/fstab. Error %s', exc)
